/* This file can be redistributed and used under GNU Library General Public
 * License. (c) 2000 Micha Moskal <malekith@topnet.pl>. 
 * $Id: utils.cc,v 1.7 2000/09/09 17:59:14 malekith Exp $ */

#define _V_PRIVATE

#include <v/config.h>
#include <v/utils.h>
#include <v/re.h>
#include <v/hash.h>

#include <stdio.h>
#include <errno.h>
#include <stdarg.h>

// ---------------------------------------------------------------------------
// Exception
// ---------------------------------------------------------------------------

Exception::Exception(const char *s)
{
	if (errno) {
		int n = errno;
		asprintf(&err_msg, "%s [errno: %s (%d)]", s, 
			 strerror(n), n);
	} else 
		err_msg = strdup(s);
	
	if (!err_msg) {
		fprintf(stderr, "\nOUT OF MEMORY during Xcpt operation\n");
		abort();
	} 
}

char *Exception::format_self()
{
	char *p;
	asprintf(&p, "%s: %s", name(), err_msg);
	return p;
}

extern int errno;

void Exception::die()
{
	fprintf(stderr, "%s\n", format_self());
	exit(1);
}

// ---------------------------------------------------------------------------
// String
// ---------------------------------------------------------------------------

String::String(const String &r)  :
	ref(strdup(r.ref))
{
	mem();
}

String::String(const char *s) :
	ref(strdup(s ? s : ""))
{
	mem();
}

String::String(const char *s, int size) :
	ref((char*)malloc(size + 1))
{
	memset(ref, ' ', size);
	ref[size] = 0;
	if (s)
		memcpy(ref, s, size);
	mem();
}

void String::mem()
{
	if (ref)
		return;
	fprintf(stderr, "i'm very sorry - out of memory during string operations...\n");
	abort();
}

void String::set(const char *x)
{
	char *tmp;
	tmp = strdup(x);
	free(ref);
	ref = tmp;
	mem();
}

void String::set(const char *x, int size)
{
	char *tmp;
	tmp = ref;
	ref = (char*)malloc(size + 1);
	mem();
	ref[size] = 0;
	memcpy(ref, x, size);
	free(tmp);
}

void String::setf(const char *fmt, ... ) 
{
	va_list ap;
	free(ref);
	va_start(ap, fmt);
	vasprintf(&ref, fmt, ap);
	va_end(ap);
	mem();
}

String xf(const char *fmt, ... ) 
{
	String r;
	
	va_list ap;
	free(r.ref);
	va_start(ap, fmt);
	vasprintf(&r.ref, fmt, ap);
	va_end(ap);
	r.mem();
	
	return r;
}

bool String::operator *=(const Regex&r)
{
	// little cheat ...
	return ((Regex *)&r)->execute(*this);
}

const String &String::operator += (const String &r)
{
	char *p = (char*)malloc(strlen() + r.strlen() + 1); 
	mem();
	strcpy(p,ref); strcat(p, r); 
	free(ref);
	ref = p;
	return *this;
}

String cnv(double d, int bd, int ad)
{
	String r1, r2;
	int c;
	int m = 1;
	
	if (d < 0) { d = -d; m = -1; }
	
	c = (int)d;
	d -= c;
	r1.setf("%.*f", ad, d);
	if (*(const char *)r1 == '1')
		c++;
	r2.setf("%*d%s", bd, c * m, (const char*)r1 + 1);
	return r2;
}

String cnv(int n, int len)
{
	String r;
	r.setf("%*d", len, n);
	return r;
}

String cnv(const String &s, int len)
{
	String r;
	if (len < 0)
		r.setf("%-*s", -len, (const char*)s);
	else
		r.setf("%*s", len, (const char*)s);
	return r;
}

// ---------------------------------------------------------------------------
// Regex
// ---------------------------------------------------------------------------

#define MAX_SUBSTR 30
Regex::Regex() :
	handle(0), pmatch(0), pattern(""), icase(false), nosub(false)
{
}

Regex::Regex(const String &p) :
	handle(0), pmatch(0), pattern(p), icase(false), nosub(false)
{
	compile();
}

Regex::Regex(const char *p) :
	handle(0), pmatch(0), pattern(p), icase(false), nosub(false)
{
	compile();
}

void Regex::compile()
{
	int err;
	if (handle) {
		regfree(handle);
		delete handle;
	}
	handle = new regex_t;
	err =
	    regcomp(handle, pattern,
		    REG_EXTENDED | REG_NEWLINE | (icase ? REG_ICASE : 0) |
		    (nosub ? REG_NOSUB : 0));
	if (err) {
		char errbuf[256];
		regerror(err, handle, errbuf, 256);
		regfree(handle);
		delete handle;
		handle = 0;
		throw XRegex(errbuf);
	}
	if (!nosub && pmatch == 0)
		pmatch = new regmatch_t[MAX_SUBSTR];
}

bool Regex::execute(const String &s)
{
	int r;
	buf = s;
	r = regexec(handle, buf, nosub ? 0 : MAX_SUBSTR, pmatch, 0);
	if (r == REG_NOMATCH)
		return false;
	else if (r == 0)
		return true;
	else
		throw XRegex("bad regexec() return code");
}

String Regex::r(int n)
{
	int len;
	
	if (n > MAX_SUBSTR)
		throw XRegex("r() arg 2big");
	if (pmatch[n].rm_so == -1)
		throw XRegex("r(): rm_so == -1");
	len = pmatch[n].rm_eo - pmatch[n].rm_so;
	if (len < 0)
		throw XRegex("len<0 !");
	if (len == 0)
		return "";
	else
		return String((const char*) buf + pmatch[n].rm_so, len);
}

Regex::~Regex()
{
	if (handle) {
		regfree(handle);
		delete handle;
	}
}

// ---------------------------------------------------------------------------
// Hash
// ---------------------------------------------------------------------------

// this probably isn't optimal...
int HashBase::get_hash(const char *str)
{
	int sum = 0xdeadf00d, len, i;
	
	if (!str || !str[0])
		return 0;
	len = strlen(str);
	
	for (i = 0; i < len; i+=2) sum += (str[i]&0x1f);
	sum *= 3;
	for (i = 1; i < len; i+=3) sum += (str[i]&0x1f);
	sum *= 5;
	for (i = 0; i < len; i+=1) sum += (str[i]&0x1f);
	return sum;
}
